#ifndef DEA_H
#define DAE_H


#ifdef __cplusplus
extern "C"
{
#endif /* __cplusplus */


/** 
    Dfinition des types de donnes.
    Les types de CPT sont abstraits.
    Trois structures sont disponibles pour crer des tableaux de fluents, de 
    prdicats et de termes, pour faciliter le passage de paramtres et le
    retour des fonctions.
**/

typedef struct Fluent * CPTFluent;
typedef struct SolutionPlan * CPTSolutionPlan;
typedef struct PDDLPredicate * CPTPredicate;
typedef struct PDDLTerm * CPTTerm;
typedef struct CPTFluentArray CPTFluentArray;
typedef struct CPTPredicateArray CPTPredicateArray;
typedef struct CPTTermArray CPTTermArray;
typedef struct CPTCostArray CPTCostArray;


struct CPTFluentArray {
  CPTFluent *elems;
  int nb;
};

struct CPTPredicateArray {
  CPTPredicate *elems;
  int nb;
};

struct CPTTermArray {
  CPTTerm *elems;
  int nb;
};

struct CPTCostArray {
  long *elems;
  int nb;
};


/**
   Initialise CPT. Le paramtre est une chane de caractres reprsentant les
   options de cpt, telles que dfinies par la ligne de commande.
**/

extern void cptMakeProblem(char *cmd_line);

/**
   Retourne le tableaux des fluents de l'tat initial.
**/

extern CPTFluentArray cptGetInit(void);

/**
   Retourne le tableaux des fluents du but.
**/

extern CPTFluentArray cptGetGoal(void);

/**
   Retourne la chane de caractres qui reprsente un fluent.
**/

extern char *cptGetFluentName(CPTFluent fluent);

/**
   Retourne l'index (unique) d'un fluent.
**/

extern int cptGetIndexFromFluent(CPTFluent fluent);

/**
   Retourne le fluent (unqiue) correspondant  un index.
**/

extern CPTFluent cptGetFluentFromIndex(int index);

/**
   Retourne le tableau des prdicats utiles du problme.
   Les prdicats utiliss uniquement pendant le grounding du problme, qui 
   n'apparaissent plus dans les actions instancies, ne sont pas fournis.
**/

extern CPTPredicateArray cptGetPredicates(void);

/**
   Retourne le prdicat d'un fluent.
**/

extern CPTPredicate cptGetPredicate(CPTFluent fluent);

/**
   Retourne le tableau des termes (objets) d'un fluent.
**/

extern CPTTermArray cptGetTerms(CPTFluent fluent);

/**
   Retourne l'arit d'un prdicat.
**/

extern int cptGetArity(CPTPredicate predicate);

/**
   Retourne le nom d'un prdicat.
**/

extern const char * cptGetPredicateName(CPTPredicate predicate);

/**
   Retourne le tableau des termes possibles pour un prdicat et une position
   dans ses paramtres. La position va de 0 (1er paramtre)  arit - 1.
**/

extern CPTTermArray cptGetPossibleTerms(CPTPredicate predicate, int position);

/**
   Retourne le nom d'un terme.
**/

extern const char * cptGetTermName(CPTTerm term);

/**
   Retourne le fluent compos du prdicat et des termes fournis en paramtre.
   Retourne NULL si le fluent n'existe pas (a peut arriver !)
**/

extern CPTFluent cptGetFluent(CPTPredicate predicate, CPTTermArray terms);

/**
   Retourne un fluent dsign par son nom.
   Peut tre utile pour faire des manip ad-hoc.
   Retourne NULL si le fluent n'existe pas.
**/

extern CPTFluent cptGetFluentFromName(char *name);

/**
   Retourne un tableau de fluents dsigns par leur nom.
   Peut tre utile pour faire des manip ad-hoc.
   Si un fluent n'existe pas, la case du tableau correspondante aura la
   valeur NULL.
**/

extern CPTFluentArray cptGetFluentArrayFromNames(char **names, int names_nb);

/**
   Retourne le tableau des fluents de l'ensemble du problme
**/

extern CPTFluentArray cptGetAllFluents(void);

/** 
    Retourne 1 si les fluents sont mutuellement exclusifs (mutex), sinon 0.
**/

extern int cptMutex(CPTFluent fluent1, CPTFluent fluent2);

/**
   Initialise la recherche pour un nouvel individu. Doit tre effectu aussi
   avant de lancer la recherche sur la premire station.
   EFFET DE BORD : dsalloue les plans intermdiaires trouvs sur l'individu
   prcdent, mais pas le plan rsultant de la compression.
**/

extern void cptResetSearch(void);

/**
   Retourne le plan solution pour une station. Ce plan sera automatiquement
   dsallou lors du prochain appel  cptResetSearch.
   Retourne NULL si le plan n'existe pas.
**/

extern CPTSolutionPlan cptSolveNext(CPTFluentArray goal);

/**
   Retourne le plan rsultant de la compression.
   En paramtre, indiquer si l'on souhaite utiliser les liens causaux des plans
   partiels (implment) et/ou les relations d'ordre (non implment).
   Ce plan n'est jamais dsallou par CPT.
   Retourne NULL si le plan n'existe pas.
**/

extern CPTSolutionPlan cptCompressSolution(bool compress_causals, bool compress_orderings);

/**
   Ecrit un plan solution dans un fichier.
**/

extern void cptPrintSolutionPlan(FILE *out, CPTSolutionPlan solution_plan);

/**
   Dsalloue un plan solution.
   Attention :  n'utiliser que sur un plan obtenu par la compression.
**/

extern void cptFreeSolutionPlan(CPTSolutionPlan solution_plan);

/**
   Retourne le makespan d'un plan solution.
**/

extern long cptGetMakespan(CPTSolutionPlan plan);

/**
   Retourne la longueur (nombre d'actions) d'un plan solution.
**/

extern int cptGetLength(CPTSolutionPlan plan);

/**
   Retourne le nombre de backtracks d'un plan solution.
**/

extern long cptGetBacktracks(CPTSolutionPlan plan);

/**
   Retourne le temps de recherche d'un plan solution.
**/

extern double cptGetSearchTime(CPTSolutionPlan plan);

/**
   Retourne le temps total pour l'obtention d'un plan.
   Si le plan rsulte d'une station intermdiaire, la valeur est identique 
   celle donne par cptGetSearchTime.
   Si le plan rsulte de la compression la valeur est la somme des temps de
   recherche des plans des stations intermdiaires plus le temps de recherche
   de la compression (i.e., c'est le temps total de calcul d'un individu.
**/

extern double cptGetTotalTime(CPTSolutionPlan plan);

/**
   Retourne le temps ncessaire au pre-processing, qui ne sera effectu qu'une 
   seule fois lors de l'appel  cptMakeProblem.
**/

extern double cptGetPreprocessingTime(CPTSolutionPlan plan);

/**
   Dfinit le nombre maximum de backtracks autoriss pour les prochains appels.
**/

extern void cptSetMaxBacktracks(long backtracks);


/**
   Donne le code de retour du dernier appel  CPT (cptSolveNext ou 
   cptCompressSolution).
**/

#define CPT_PLAN_FOUND 0	// on a trouv une solution
#define CPT_GOALS_MUTEX 1	// les buts sont mutex
#define CPT_INIT_PROP_FAILED 2	// on choue dans la propagation initiale
#define CPT_BACKTRACK_LIMIT 3	// la limite de backtracks est atteinte
#define CPT_BOUND_LIMIT 4	// il n'y a pas de plan pour un makespan maximum dfini par l'option -m
#define CPT_NO_PLAN 5		// il n'y a pas de plan (yahsp)

extern int cptGetReturnCode(void);

/**
   Retourne le tableau des cots possibles des fluents
**/

extern CPTCostArray cptGetCosts();

/**
   Retourne le tableau des fluents dont le cot est pass en paramtre
**/

extern CPTFluentArray cptGetFluentsByCost(long cost);

/**
   Retourne un entier (1 octet) reprsentant le pourcentage de couples d'actions
   mutex produisant un couple de fluents. Si f1 == f2, le rsultat est 0. Si f1
   et f2 sont mutex, le rsultat est 100.
**/

extern char cptGetFluentDifficulty(CPTFluent f1, CPTFluent f2);

/**
   Retourne le pgcd des dures
**/

extern long cptGetPGCD();

/**
   Retourne le ppcm des dures
**/

extern long cptGetPPCM();

/**
   Initialise CPT (pas encore YAHSP) pour fonctionner en mode PDDL2.1.
   Les paramtre est la prcision demande sous forme de chane de caractres,
   par exemple "0.001".
**/

extern void cptSetPDDL21(char *precision);

/** 
    Limite la propagation initiale  max_propagations.
**/

  extern void cptSetPropagationLimit(long max_propagations);

#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* DAE_H */
